import pytest
import numpy as np

from cosapp.systems import System
from cosapp.multimode import Event
from cosapp.recorders import DataFrameRecorder
from cosapp.drivers import RungeKutta
from .conftest import ScalarOde


class ChildFSystem(System):
    def setup(self):
        self.add_inward("x")
        input = self.add_event("input_event")
        self.add_event("e", trigger = input.filter("x >= -1"))

    def transition(self):
        if self.e.present:
            self.x = -self.x


class TopFSystem(System):
    def setup(self):
        self.add_child(ScalarOde("ode"), pulling=['f', 'df'])
        left = self.add_child(ChildFSystem("left"))
        right = self.add_child(ChildFSystem("right"))
        p = self.add_event("p", trigger = "f >= 1.")
        left.input_event.trigger = p
        right.input_event.trigger = p

    def transition(self):
        if self.p.present:
            self.f = 0.


def test_MultimodeSystem_filter():

    period = 0.15
    top = TopFSystem("top")
    driver = top.add_driver(RungeKutta("driver", order = 3, time_interval = (0, 1.5), dt = period))
    driver.add_recorder(DataFrameRecorder(includes=['f', 'left.x', 'right.x']), period = period)

    driver.set_scenario(
        init = {'f': 0., 'left.x': 1.5, 'right.x': -0.5},
        values = {'df': '2. * t'},
    )

    top.run_drivers()
    data = driver.recorder.export_data()

    te1 = 1.
    te2 = np.sqrt(2.)
    exact_t = np.array([0, 0.15, 0.3, 0.45, 0.6, 0.75, 0.9, te1, te1, 1.05, 1.2, 1.35, te2, te2, 1.5])
    exact_f = [0, 0.0225, 0.09, 0.2025, 0.36, 0.5625, 0.81, 1, 0, 0.1025, 0.44, 0.8225, 1, 0, 0.25]
    exact_left_x = [1.5] * 8 + [-1.5] * 7
    exact_right_x = [-0.5] * 8 + [0.5] * 5 + [-0.5] * 2
    expected = {
        'f': exact_f,
        'left.x': exact_left_x,
        'right.x': exact_right_x,
        'time': exact_t
    }
    assert np.asarray(data['time']) == pytest.approx(np.asarray(expected['time']), abs=1.e-12)
    assert np.asarray(data['f']) == pytest.approx(expected['f'], abs=1.e-12)
    assert np.asarray(data['left.x']) == pytest.approx(expected['left.x'], abs=1.e-12)
    assert np.asarray(data['right.x']) == pytest.approx(expected['right.x'], abs=1.e-12)


class CascadingSystem(System):
    def setup(self):

        self.add_inward("x1", 0.5)
        self.add_inward("x2", 1.5)
        self.add_inward("x3", 2.5)

        self.add_child(ScalarOde("ode"), pulling=['f', 'df'])

        p = self.add_event("p", trigger = "f >= 0.863")
        p1 = self.add_event("p1", trigger = "x1 < 0.")
        p2 = self.add_event("p2", trigger = "x2 < 0.")
        p3 = self.add_event("p3", trigger = "x3 < 0.")

        self.add_event("merged_event", trigger = Event.merge(p, p1, p2, p3))

    def transition(self):
        if self.merged_event.present:
            self.x1 -= 1.
            self.x2 -= 1.
            self.x3 -= 1.


def test_MultimodeSystem_cascade():
    period = 0.1
    top = CascadingSystem("top")
    driver = top.add_driver(RungeKutta("driver", order = 3, time_interval = (0, 1), dt = period))
    driver.add_recorder(DataFrameRecorder(includes=['f', 'x1', 'x2', 'x3']), period = period)

    driver.set_scenario(
        init = {'f': 0., 'x1': 0.5, 'x2': 1.5, 'x3': 2.5},
        values = {'df': 1.},
    )

    top.run_drivers()
    data = driver.recorder.export_data()
    assert len(driver.event_data) == 5

    te = 0.863
    exact_t = [0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8] + [te] * 2 + [0.9, 1.]
    exact_f = exact_t
    exact_x1 = [0.5] * 10 + [-3.5] * 3
    exact_x2 = [1.5] * 10 + [-2.5] * 3
    exact_x3 = [2.5] * 10 + [-1.5] * 3
    expected = {
        'f': exact_f,
        'x1': exact_x1,
        'x2': exact_x2,
        'x3': exact_x3,
        'time': exact_t
    }
    assert np.asarray(data['time']) == pytest.approx(expected['time'], abs=1.e-12)
    assert np.asarray(data['f']) == pytest.approx(expected['f'], abs=1.e-12)
    assert np.asarray(data['x1']) == pytest.approx(expected['x1'], abs=1.e-12)
    assert np.asarray(data['x2']) == pytest.approx(expected['x2'], abs=1.e-12)
    assert np.asarray(data['x3']) == pytest.approx(expected['x3'], abs=1.e-12)


class CascadingParentSystem(System):
    def setup(self):
        self.add_child(ScalarOde("ode"), pulling = ['f', 'df'])
        self.add_event("p", trigger = "f >= 0.863")
        self.add_event("merged_event")


class CascadingChildSystem(System):
    def setup(self):
        self.add_inward("x")
        self.add_event("p", trigger = "x < 0.")
        self.add_event("m")

    def transition(self):
        if self.m.present:
            self.x -= 1.


def test_MultimodeSystem_cascade2():
    period = 0.1
    top = CascadingParentSystem("top")
    c1 = top.add_child(CascadingChildSystem("c1"))
    c2 = top.add_child(CascadingChildSystem("c2"))
    c3 = top.add_child(CascadingChildSystem("c3"))

    top.merged_event.trigger = Event.merge(top.p, c1.p, c2.p, c3.p)
    c1.m.trigger = top.merged_event
    c2.m.trigger = top.merged_event
    c3.m.trigger = top.merged_event

    driver = top.add_driver(RungeKutta("driver", order=3, time_interval=(0, 1), dt=period))
    driver.add_recorder(DataFrameRecorder(includes=['f', 'c1.x', 'c2.x', 'c3.x']), period=period)

    driver.set_scenario(
        init = {'f': 0., 'c1.x': 0.5, 'c2.x': 1.5, 'c3.x': 2.5},
        values = {'df': 1.},
    )

    top.run_drivers()
    data = driver.recorder.export_data()
    assert len(driver.event_data) == 5
    #print(driver.event_data)

    te = 0.863
    exact_t = [0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8] + [te] * 2 + [0.9, 1.]
    exact_f = exact_t
    exact_x1 = [0.5] * 10 + [-3.5] * 3
    exact_x2 = [1.5] * 10 + [-2.5] * 3
    exact_x3 = [2.5] * 10 + [-1.5] * 3
    expected = {
        'f': exact_f,
        'x1': exact_x1,
        'x2': exact_x2,
        'x3': exact_x3,
        'time': exact_t
    }
    assert np.asarray(data['f']) == pytest.approx(expected['f'], abs=1.e-12)
    assert np.asarray(data['time']) == pytest.approx(expected['time'], abs=1.e-12)
    assert np.asarray(data['c1.x']) == pytest.approx(expected['x1'], abs=1.e-12)
    assert np.asarray(data['c2.x']) == pytest.approx(expected['x2'], abs=1.e-12)
    assert np.asarray(data['c3.x']) == pytest.approx(expected['x3'], abs=1.e-12)

class LongCascadeSystem(System):
    def setup(self):

        # Primal time event, to be triggered a second time during the cascade
        # because of resets of f during certain transitions
        self.add_child(ScalarOde("ode"), pulling = ['f', 'df'])
        e_f = self.add_event("e_t", trigger = "f == 0.42")

        # Ad hoc variables for ensuring a cascade of events
        self.add_outward("x", np.array([-0.5,-1.5,-2.5,-3.5,-4.5]))
        e_0 = self.add_event("e_0", trigger = "x[0] > 0")
        e_1 = self.add_event("e_1", trigger = "x[1] > 0")
        e_2 = self.add_event("e_2", trigger = "x[2] > 0")
        e_3 = self.add_event("e_3", trigger = "x[3] > 0")
        e_4 = self.add_event("e_4", trigger = "x[4] > 0")

        # Merged event: all x_i's are changed at each discrete step
        self.add_event("any_event", trigger = Event.merge(e_f, e_0, e_1, e_2, e_3, e_4))

    def transition(self):
        if self.any_event.present:
            self.x += np.ones(5)
        if self.e_1.present: # e_f only unlocks after this transition
            self.f = -1.
        if self.e_4.present: # e_f triggers once again after this transition
            self.f = 1.

def test_MultimodeSystem_cascade_event_lock():
    """The event locking mechanism consists in preventing zero-crossing (ZC) events to erroneously trigger
    in rapid succession. Among others, it handles the case of bidirectional ZC events, which
    would be detected twice instead of only once without this mechanism: first when the ZC function
    reaches 0, and then, as soon as its value drifts away from 0.
    
    In this test, the first triggered event is top.e_f, interrupting an integration time step
    at time t=0.42. Then, events top.e_0 to top.e_4, in this order, will trigger successive
    discrete steps.
    
    The transition caused by e_1 makes the value of f switch from 0 to -1; this would cause event top.e_f
    to immediately trigger if the event locking mechanism was not (or incorrectly) implemented.
    On the other hand, as event top.e_f has to be unlocked just after this step, it has to trigger after
    the discrete step caused by top.e_4, as the associated transition makes the value of f switch from -1 to 1.

    As a result, a discrete loop of exactly seven discrete steps, each triggered by a single primitive event
    (e_f, e_0, e_1, e_2, e_3, e_4, e_f in this order), is expected.
    """
    period = 0.1
    top = LongCascadeSystem("top")

    driver = top.add_driver(RungeKutta("driver", order=3, time_interval=(0, 0.6), dt=period))
    driver.add_recorder(DataFrameRecorder(includes=['f', 'x']), period=period)

    driver.set_scenario(
        init = {'f': 0.},
        values = {'df': 1.},
    )

    top.run_drivers()
    data = driver.recorder.export_data()
    assert len(driver.event_data) == 8  # A single discrete loop of 7 steps

    expected = {
        'f': [0, 0.1, 0.2, 0.3, 0.4, 0.42, 1, 1.08, 1.18],
        'time': [0, 0.1, 0.2, 0.3, 0.4, 0.42, 0.42, 0.5, 0.6],
        'x': [
            [-0.5] * 6 + [6.5] * 3,  # time values of x[0]
            [-1.5] * 6 + [5.5] * 3,  # time values of x[1]
            [-2.5] * 6 + [4.5] * 3,  # etc.
            [-3.5] * 6 + [3.5] * 3,
            [-4.5] * 6 + [2.5] * 3,
        ],
    }

    assert np.asarray(data['f']) == pytest.approx(expected['f'], abs=1.e-12)
    assert np.asarray(data['time']) == pytest.approx(expected['time'], abs=1.e-12)
    
    # Transpose data such that x[i] contains time values of top.x[i]
    x = np.transpose(data['x'].tolist())

    for i, xi in enumerate(x):
        assert xi == pytest.approx(expected['x'][i], abs=1e-12), f"@ i = {i}"



import copy


def get_state(system: "System") -> dict:
    """Export system state (inputs and outputs) into a dictionary
    for quick serialization, without type check.
    
    Returns:
    --------
    Recursive dictionary of the kind:
    {
        'ports: {
            port_name: {
                varname: value,
                ...
            },
            ...
        },
        'children': {
            child_name: {
                'ports': {...},
                'children: {...},
            },
            ...
        },
    }
    """
    from cosapp.base import System
    def recursive_search(system: "System", state: dict) -> None:
        state['ports'] = port_data = {}
        for port in system.ports():
            data = {
                var: copy.deepcopy(value)
                for var, value in port.items()
            }
            if data:
                port_data[port.name] = data
        if system.children:
            state['children'] = child_data = {}
            for child in system.children.values():
                child_data[child.name] = child_state = {}
                recursive_search(child, child_state)
    state = {}
    recursive_search(system, state)
    return state


def set_state(system: "System", state: dict) -> None:
    """Import system state (inputs and outputs) from
    a dictionary generated by `get_state`.
    """
    for name, values in state['ports'].items():
        port = getattr(system, name)
        for name, value in values.items():
            setattr(port, name, copy.deepcopy(value))
        # system[name].set_values(**copy.deepcopy(values))
    child_data = state.get('children', {})
    for name, child_state in child_data.items():
        set_state(system[name], child_state)



import copy
import numpy
import abc
import csv
from io import StringIO
from numbers import Number
from typing import (
    Any, Callable, Dict, List, Optional,
    Sequence, Tuple, Union, Iterable,
    TypeVar,
)

from cosapp.core.numerics.basics import MathematicalProblem, SolverResults
from cosapp.core.numerics.enum import NonLinearMethods
from cosapp.core.numerics.root import root
from cosapp.drivers.driver import Driver
from cosapp.drivers.abstractsolver import AbstractSolver
from cosapp.drivers.runsinglecase import RunSingleCase
from cosapp.drivers.utils import DesignProblemHandler
from cosapp.utils.helpers import check_arg
from cosapp.utils.logging import LogFormat, LogLevel

import logging
logger = logging.getLogger(__name__)

AnyDriver = TypeVar("AnyDriver", bound=Driver)


class BaseSolverBuilder(abc.ABC):
    """Base interface for numerical system building strategy used by `NonLinearSolver`.
    Implementations will differ for single- or multi-point design.
    """
    def __init__(self, solver: AbstractSolver):
        self.solver = solver
        self.problem = MathematicalProblem(solver.name, solver.owner)
        self.initial_values = numpy.empty(0)
        self.is_design_unknown: Dict[str, bool] = dict()

    @abc.abstractmethod
    def build_system(self) -> None:
        pass

    @abc.abstractmethod
    def update_residues(self) -> None:
        pass

    def update_unknowns(self, x: numpy.ndarray) -> None:
        counter = 0
        for name, unknown in self.problem.unknowns.items():
            if unknown.mask is None:
                unknown.set_default_value(x[counter])
                counter += 1
            else:
                n = numpy.count_nonzero(unknown.mask)
                unknown.set_default_value(x[counter : counter + n])
                counter += n
            # Update design unknowns
            if self.is_design_unknown[name]:
                unknown.set_to_default()


class NonLinearSolver(AbstractSolver):
    """Solve mathematical problem with algebraic variables.

    Attributes
    ----------
    compute_jacobian : bool
        Should the Jacobian matrix be computed? ; default True
    jac_lup : ndarray, optional
        LU decomposition of latest Jacobian matrix (if available, None otherwise)
    jac : ndarray, optional
        Latest Jacobian matrix computed (if available, None otherwise)
    """

    __slots__ = (
        '__method', '__option_aliases', '__trace', '__results',
        '__design_unknowns', 'compute_jacobian', 'jac_lup', 'jac',
        '__builder',
    )

    def __init__(self, 
        name: str, 
        owner: Optional["cosapp.systems.System"] = None, 
        method: Union[NonLinearMethods, str] = NonLinearMethods.NR, 
        **kwargs
    ) -> None:
        """Initialize driver

        Parameters
        ----------
        name: str, optional
            Name of the `Driver`.
        owner: System, optional
            :py:class:`~cosapp.systems.system.System` to which this driver belong; defaults to `None`.
        method : Union[NonLinearMethods, str]
            Resolution method to use
        **kwargs:
            Additional keywords arguments forwarded to base class.
        """
        super().__init__(name, owner, **kwargs)

        if isinstance(method, str):
            method = NonLinearMethods[method]
        else:
            check_arg(method, 'method', NonLinearMethods)
        self.__method = method
        self.__option_aliases = dict()
        self.__set_method(method, **kwargs)
        self.__trace: List[Dict[str, Any]] = list()
        self.__results: SolverResults = None
        self.__builder: BaseSolverBuilder = None

        self.compute_jacobian = True  # type: bool
            # desc='Should the Jacobian matrix be computed?'

        self.jac_lup = (None, None)  # type: Tuple[Optional[numpy.ndarray], Optional[numpy.ndarray]]
            # desc='LU decomposition of latest Jacobian matrix (if available).'
        self.jac = None  # type: Optional[numpy.ndarray]
            # desc='Latest computed Jacobian matrix (if available).'

    def reset_problem(self) -> None:
        """Reset mathematical problem"""
        super().reset_problem()
        self.__design_unknowns = dict()

    @property
    def method(self) -> NonLinearMethods:
        """NonLinearMethods : Selected solver algorithm."""
        return self.__method

    @property
    def results(self) -> SolverResults:
        """SolverResults: structure containing solver results,
        together with additional detail.
        """
        return self.__results

    def add_child(self, child: AnyDriver, execution_index: Optional[int]=None, desc="") -> AnyDriver:
        """Add a child `Driver` to the current `Driver`.

        When adding a child `Driver`, it is possible to specified its position in the execution order.

        Child `Port`, `inwards` and `outwards` can also be pulled at the parent level by providing either
        the name of the port/inward/outward or a list of them or the name mapping of the child element
        (dictionary keys) to the parent element (dictionary values).
        If the argument is not a dictionary, the name in the parent system will be the same as in the child.

        Parameters
        ----------
        - child: Driver
            `Driver` to add to the current `Driver`
        - execution_index: int, optional
            Index of the execution order list at which the `Module` should be inserted;
            default latest.
        - desc [str, optional]:
            Sub-driver description in the context of its parent driver.

        Notes
        -----
        The added child will have its owner set to match the one of the current driver.
        """
        self.compute_jacobian = True
        return super().add_child(child, execution_index, desc)

    def is_standalone(self) -> bool:
        """Is this Driver able to solve a system?

        Returns
        -------
        bool
            Ability to solve a system or not.
        """
        return True

    def resolution_method(self,
        fresidues: Callable[[Sequence[float], Union[float, str], bool], numpy.ndarray],
        x0: Sequence[float],
        args: Tuple[Union[float, str]] = (),
        options: Optional[Dict[str, Any]] = None
    ) -> SolverResults:

        if self.method == NonLinearMethods.NR:
            self.options.update(self._get_solver_limits())

        common_keys = set(self.options).intersection(options)
        this_options = dict((key, options[key]) for key in common_keys)

        if self.method == NonLinearMethods.NR:
            this_options['compute_jacobian'] = self.compute_jacobian
            this_options['jac_lup'] = self.jac_lup
            this_options['jac'] = self.jac

        results = root(fresidues, x0, args=args, method=self.method, options=this_options)

        if results.jac_lup[0] is not None:
            self.jac_lup = results.jac_lup
            self.jac = results.jac
            self.compute_jacobian = False

        return results

    def setup_run(self) -> None:
        super().setup_run()
        self._init_problem()

    def _print_solution(self) -> None:  # TODO better returning a string
        """Print the solution in the log."""
        if self.options['verbose']:
            # TODO move it in MathematicalProblem
            logger.info(f"Parameters [{len(self.solution)}]: ")
            for k, v in self.solution.items():
                logger.info(f"   # {k}: {v}")

            for name, residue in self.problem.residues.items():
                value = numpy.asarray(residue.value)
                if value.ndim > 0:
                    spacing = " " * len(name)
                    msg = f"Residues [{len(self.problem.residues)}]: \n   # ({name}"
                    msg += f"\n   #  {spacing}".join(f", {v:.5g}" for v in value)
                    msg += ")\n"
                else:
                    logger.info(f"   # ({name}, {value:.5g})")

            tol = numpy.linalg.norm(self.problem.residue_vector(), numpy.inf)
            try:
                option = self.__option_aliases['tol']  # should never fail, in principle
                target = self.options[option]
            except:
                target = 0
            logger.debug(f" # Current tolerance {tol} for target {target}")

    def _init_problem(self):
        """Initialize mathematical problem"""
        logger.debug(
            "\n".join([
                "*" * 40,
                "*", 
                "* Assemble mathematical problem",
                "*",
                "*" * 40,
            ])
        )
        run_cases = list(
            filter(
                lambda driver: isinstance(driver, RunSingleCase),
                self.children.values(),
            )
        )
        if run_cases:
            builder = MultipointSolverBuilder(self, run_cases)
        else:
            builder = StandaloneSolverBuilder(self)
        
        builder.build_system()
        self.problem = builder.problem
        self.initial_values = builder.initial_values
        self.__builder = builder
        self.touch_unknowns()
        logger.debug(
            "\n".join([
                "Mathematical problem:",
                f"{'<empty>' if self.problem.is_empty() else self.problem}",
            ])
        )

    def _fresidues(self, x: Sequence[float]) -> numpy.ndarray:
        """
        Method used by the solver to take free variables values as input and values of residues as
        output (after running the System).

        Parameters
        ----------
        x : Sequence[float]
            The list of values to set to the free variables of the `System`
        update_residues_ref : bool
            Request residues to update their reference

        Returns
        -------
        numpy.ndarray
            The list of residues of the `System`
        """
        x = numpy.asarray(x)
        logger.debug(f"Call fresidues with x = {x!r}")
        self.set_iteratives(x)

        if self.children:
            for subdriver in self.children.values():
                logger.debug(f"Call {subdriver.name}.run_once()")
                subdriver.run_once()
        else:
            self.owner.run_children_drivers()

        self.__builder.update_residues()
        residues = self.problem.residue_vector()
        logger.debug(f"Residues: {residues!r}")
        return residues

    def set_iteratives(self, x: Sequence[float]) -> None:
        x = numpy.asarray(x)
        self.__builder.update_unknowns(x)

    def compute(self) -> None:
        """Run the resolution method to find free vars values that zero out residues
        """
        # Reset status
        self.status = ''
        self.error_code = '0'

        try:
            self.problem.validate()
        except ArithmeticError:
            self.status = 'ERROR'
            self.error_code = '9'
            raise

        if len(self.initial_values) > 0:
            self.solution = {}

            # compute first order
            results = self.resolution_method(
                self._fresidues,
                self.initial_values,
                options=self.options,
            )

            self.__results = results
            self.__trace = getattr(results, "trace", list())

            if results.success:
                self.status = ''
                self.error_code = '0'
                logger.info(f"solver : {self.name}{results.message}")
            
            else:
                self.status = 'ERROR'
                self.error_code = '9'

                error_msg = f"The solver failed: {results.message}"

                error_desc = getattr(results, 'jac_errors', {})
                if error_desc:
                    if (indices := error_desc.get('unknowns', [])):
                        unknown_names = self.problem.unknown_names()
                        error_msg += (
                            f" \nThe {len(indices)} following parameter(s)"
                            f" have no influence: {[unknown_names[i] for i in indices]} \n{indices}"
                        )
                    if (indices := error_desc.get('residues', [])):
                        residue_names = self.problem.residue_names()
                        error_msg += (
                            f" \nThe {len(indices)} following residue(s)"
                            f" are not influenced: {[residue_names[i] for i in indices]}"
                        )

                if self.parent is not None:
                    raise ArithmeticError(error_msg)
                else:
                    logger.error(error_msg)

            self.solution = dict(
                (key, unknown.default_value)
                for key, unknown in self.problem.unknowns.items()
            )
            self._print_solution()

        else:
            logger.debug('No parameters/residues to solve. Fallback to children execution.')
            for child in self.children.values():
                child.run_once()
            self.owner.run_children_drivers()

        if self._recorder is not None:
            if self.children:
                for child in self.children.values():
                    child.run_once()
                    self._recorder.record_state(child.name, self.status, self.error_code)
            else:
                self._recorder.record_state(self.name, self.status, self.error_code)

    def log_debug_message(self,
        handler: "HandlerWithContextFilters",
        record: logging.LogRecord,
        format: LogFormat = LogFormat.RAW
    ) -> bool:
        """Callback method on the driver to log more detailed information.
        
        This method will be called by the log handler when :py:meth:`~cosapp.utils.logging.LoggerContext.log_context`
        is active if the logging level is lower or equals to VERBOSE_LEVEL. It allows
        the object to send additional log message to help debugging a simulation.

        Parameters
        ----------
        handler : HandlerWithContextFilters
            Log handler on which additional message should be published.
        record : logging.LogRecord
            Log record
        format : LogFormat
            Format of the message

        Returns
        -------
        bool
            Should the provided record be logged?
        """
        message = record.getMessage()
        activate = getattr(record, "activate", None)
        emit_record = super().log_debug_message(handler, record, format)

        try:
            self.options["history"] = activate
        except (KeyError, TypeError):
            pass

        if message.endswith("call_setup_run") or message.endswith("call_clean_run"):
            emit_record = False

        elif activate == True:
            emit_record = False

        elif activate == False:
            emit_record = False

            message = ""
            unknown_trace = None
            residue_trace = None
            residue_names = self.problem.residue_names()
            unknown_names = self.problem.unknown_names()
            for i, record in enumerate(self.__trace):
                if i == 0:
                    unknown_trace = record["x"]
                    residue_trace = record["residues"]
                else:
                    unknown_trace = numpy.vstack((unknown_trace, record["x"]))
                    residue_trace = numpy.vstack((residue_trace, record["residues"]))
                try:
                    jac_record = record["jac"]
                except KeyError:
                    continue
                else:
                    message += f"Iteration {i}\n"
                    with StringIO() as container:
                        numpy.savetxt(container, jac_record, delimiter=",")
                        jacobian = container.getvalue()
                    unknowns = ", ".join(unknown_names)
                    message += f"New Jacobian matrix:\n,{unknowns}\n"
                    for residue, line in zip(residue_names, jacobian.splitlines()):
                        message += f"{residue}, {line}\n"

            def format_record(records: numpy.ndarray, headers: List[str]) -> str:
                records = numpy.atleast_2d(records)
                with StringIO() as stream:
                    writer = csv.writer(stream, delimiter=",", lineterminator="\n")
                    writer.writerows([headers, *records])
                    return stream.getvalue()

            if unknown_trace is not None:
                message += "Unknowns\n{}\n".format(
                    format_record(unknown_trace, unknown_names)
                )
            if residue_trace is not None:
                message += "Residues\n{}\n".format(
                    format_record(residue_trace, residue_names)
                )

            handler.log(
                LogLevel.FULL_DEBUG,
                message,
                name=logger.name,
            )

        return emit_record

    def __set_method(self, method, **options):
        check_arg(method, 'method', NonLinearMethods)

        self.__method = method
        self.__option_aliases = dict()

        if self.__method == NonLinearMethods.NR:
            self.options.declare('tol', 'auto', dtype=(float, str), allow_none=True,
                desc='Absolute tolerance (in max-norm) for the residual.')
            self.options.declare('max_iter', 500, dtype=int,
                desc='The maximum number of iterations.')
            # Note: use a power of 2 for `eps`, to guaranty machine-precision accurate gradients in linear problems
            self.options.declare('eps', 2**(-16), dtype=float, allow_none=True, lower=2**(-30),
                desc='Relative step length for the forward-difference approximation of the Jacobian.')
            self.options.declare('factor', 1.0, dtype=Number, allow_none=True, lower=1e-3, upper=1.0,
                desc='A parameter determining the initial step bound factor * norm(diag * x). Should be in interval [0.001, 1].')
            self.options.declare('partial_jac', True, dtype=bool, allow_none=False,
                desc='Defines if partial Jacobian updates can be computed before a complete Jacobian matrix update.')
            self.options.declare('partial_jac_tries', 10, dtype=int, allow_none=False, lower=1, upper=10,
                desc='Defines how many partial Jacobian updates can be tried before a complete Jacobian matrix update.')
            self.options.declare('jac_update_tol', 0.01, dtype=float, allow_none=False, lower=0, upper=1,
                desc='Tolerance level for partial Jacobian matrix update, based on nonlinearity estimation.')
            self.options.declare('recorder', None, allow_none=True,
                desc='A recorder to store solver intermediate results.')
            self.options.declare('lower_bound', None, dtype=numpy.ndarray, allow_none=True,
                desc='Min values for parameters iterated by solver.')
            self.options.declare('upper_bound', None, dtype=numpy.ndarray, allow_none=True,
                desc='Max values for parameters iterated by solver.')
            self.options.declare('abs_step', None, dtype=numpy.ndarray, allow_none=True,
                desc='Max absolute step for parameters iterated by solver.')
            self.options.declare('rel_step', None, dtype=numpy.ndarray, allow_none=True,
                desc='Max relative step for parameters iterated by solver.')
            self.options.declare('history', False, dtype=bool, allow_none=False,
                desc='Request saving the resolution trace.')
            self.options.declare('tol_update_period', 4, dtype=int, lower=1, allow_none=False,
                desc="Tolerance update period, in iteration number, when tol='auto'.")
            self.options.declare('tol_to_noise_ratio', 16, dtype=Number, lower=1.0, allow_none=False,
                desc="Tolerance-to-noise ratio, when tol='auto'.")

        elif self.__method == NonLinearMethods.POWELL:
            self.__option_aliases = {
                'tol': 'xtol',
                'max_eval': 'maxfev',
            }
            self.options.declare('xtol', 1.0e-7, dtype=float,
                desc="The calculation will terminate if the relative error between two consecutive iterations is at most tol.")
            self.options.declare('maxfev', 0, dtype=int,
                desc='The maximum number of calls to the function. If zero, assumes 100 * (N + 1),'
                    ' where N is the number of elements in x0.')
            self.options.declare('eps', None, dtype=float, allow_none=True,
                desc='A suitable step length for the forward-difference approximation of the Jacobian (for fprime=None).'
                    ' If eps is less than machine precision u, it is assumed that the relative errors in the'
                    ' functions are of the order of u.')
            self.options.declare('factor', 0.1, dtype=float, lower=0.1, upper=100.,
                desc='A parameter determining the initial step bound factor * norm(diag * x). Should be in the interval [0.1, 100].')

        elif self.__method == NonLinearMethods.BROYDEN_GOOD:
            self.__option_aliases = {
                'tol': 'fatol',
                'num_iter': 'nit',
                'max_iter': 'maxiter',
                'min_rel_step': 'xtol',
                'min_abs_step': 'xatol',
            }
            self.options.declare('nit', 100, dtype=int,
                desc='Number of iterations to perform. If omitted (default), iterate unit tolerance is met.')
            self.options.declare('maxiter', 200, dtype=int,
                desc='Maximum number of iterations to make. If more are needed to meet convergence, NoConvergence is raised.')
            self.options.declare('disp', False, dtype=bool,
                desc='Print status to stdout on every iteration.')
            self.options.declare('fatol', 6e-6, dtype=float,
                desc='Absolute tolerance (in max-norm) for the residual. If omitted, default is 6e-6.')
            self.options.declare('line_search', 'armijo', dtype=str, allow_none=True,
                desc='Which type of a line search to use to determine the step '
                    'size in the direction given by the Jacobian approximation. Defaults to ‘armijo’.')
            self.options.declare('jac_options', {'reduction_method': 'svd'},
                dtype=dict, allow_none=True,
                desc='Options for the respective Jacobian approximation. restart, simple or svd')

        self._filter_options(options, self.__option_aliases)

    def extend(self, problem: MathematicalProblem, *args, **kwargs) -> MathematicalProblem:
        """Extend solver inner problem.
        
        Parameters
        ----------
        - problem [MathematicalProblem]:
            Source mathematical problem.
        - *args, **kwargs:
            Additional arguments forwarded to `MathematicalProblem.extend`.

        Returns
        -------
        - MathematicalProblem:
            The extended problem.
        """
        return self._raw_problem.extend(problem, *args, **kwargs)

    def add_unknown(self,
        name: Union[str, Iterable[Union[dict, str]]],
        *args, **kwargs,
    ) -> MathematicalProblem:
        """Add design unknown(s).

        More details in `cosapp.core.MathematicalProblem.add_unknown`.

        Parameters
        ----------
        - name [str or Iterable of dictionary or str]:
            Name of the variable or collection of variables to be added.
        - *args, **kwargs:
            Additional arguments forwarded to `MathematicalProblem.add_unknown`.

        Returns
        -------
        - MathematicalProblem:
            The updated problem.
        """
        return self._raw_problem.add_unknown(name, *args, **kwargs)

    def add_equation(self,
        equation: Union[str, Iterable[Union[dict, str]]],
        *args, **kwargs,
    ) -> MathematicalProblem:
        """Add off-design equation(s).

        More details in `cosapp.core.MathematicalProblem.add_equation`.

        Parameters
        ----------
        - equation [str or Iterable of str of the kind 'lhs == rhs']:
            Equation or collection of equations to be added.
        - *args, **kwargs:
            Additional arguments forwarded to `MathematicalProblem.add_equation`.

        Returns
        -------
        - MathematicalProblem:
            The updated problem.
        """
        return self._raw_problem.add_equation(equation, *args, **kwargs)

    def add_target(self,
        expression: Union[str, Iterable[str]],
        *args, **kwargs,
    ) -> MathematicalProblem:
        """Add deferred off-design equation(s).

        More details in `cosapp.core.MathematicalProblem.add_target`.

        Parameters
        ----------
        - expression: str
            Targetted expression
        - *args, **kwargs : Forwarded to `MathematicalProblem.add_target`

        Returns
        -------
        MathematicalProblem
            The modified mathematical problem
        """
        return self._raw_problem.add_target(expression, *args, **kwargs)


class StandaloneSolverBuilder(BaseSolverBuilder):
    """System building strategy for solvers with no
    `RunSingleCase` sub-drivers. In this case, the actual
    mathematical problem is entirely defined by unknowns and
    equations declared at solver level.
    """
    def build_system(self):
        solver = self.solver
        system = self.solver.owner
        problem = self.problem
        handler = DesignProblemHandler(system)
        handler.design.extend(solver.raw_problem)
        handler.offdesign.extend(system.assembled_problem())
        handler.prune()
        problem.clear()
        problem.extend(handler.merged_problem(), copy=False)
        self.initial_values = problem.unknown_vector()
        self.is_design_unknown = dict.fromkeys(problem.unknowns, True)

    def update_residues(self):
        for residue in self.problem.residues.values():
            residue.update()


class MultipointSolverBuilder(BaseSolverBuilder):
    """System building strategy for solvers with one or more
    `RunSingleCase` sub-drivers. In this case, the actual
    mathematical problem is a combination of unknowns and
    equations declared at solver level, and of local design
    and off-design problems declared at case level.
    """
    def __init__(self, solver: NonLinearSolver, points: List[RunSingleCase]):
        super().__init__(solver)
        self.points = points

    def build_system(self):
        solver = self.solver
        system = self.solver.owner
        problem = self.problem
        handler = DesignProblemHandler(system)
        handler.design.extend(solver.raw_problem, equations=False)
        handler.offdesign.extend(solver.raw_problem, unknowns=False)
        handler.prune()  # resolve aliasing
        # Create assembled problem
        problem.clear()
        problem.extend(handler.design)

        self.initial_values = handler.design.unknown_vector()

        if len(self.points) > 1:
            # If more than one `RunSingleCase` drivers are present,
            # use a name wrapper to distinguish dict entries for
            # residues and off-design unknowns in global problem
            get_key_wrapper = lambda case: (
               lambda key: f"{case.name}[{key}]"
            )
        else:
            get_key_wrapper = lambda case: None

        # Set of unknown names to detect design/off-design conflicts
        design_unknown_names = set(handler.design.unknowns)

        for point in self.points:
            local = point.processed_problems
            # Check that local problem does not introduce design/off-design conflicts
            design_unknown_names.update(local.design.unknowns)
            common = design_unknown_names.intersection(local.offdesign.unknowns)
            if common:
                kind = "unknown"
                if len(common) > 1:
                    names = ", ".join(map(repr, sorted(common)))
                    names = f"({names}) are"
                    kind += "s"
                else:
                    names = f"{common.pop()!r} is"
                raise ValueError(
                    f"{names} defined as design and off-design {kind} in {point.name!r}"
                )
            # Enforce solver-level off-design problem to child case
            point.add_offdesign_problem(handler.offdesign)
            # Assemble case problems (design & off-design)
            wrapper = get_key_wrapper(point)
            problem.extend(local.design, copy=False, residue_wrapper=wrapper)
            problem.extend(local.offdesign, copy=False, unknown_wrapper=wrapper, residue_wrapper=wrapper)
            self.initial_values = numpy.append(self.initial_values, point.get_init(solver.force_init))

        self.is_design_unknown = dict.fromkeys(problem.unknowns, False)
        for name in design_unknown_names:
            self.is_design_unknown[name] = True

    def update_residues(self):
        # Nothing to do, since residues are updated
        # by `RunSingleCase` sub-drivers
        pass



"""
Core package of CoSApp.
"""
from cosapp.core._version import __version__

from cosapp.core.config import CoSAppConfiguration
from cosapp.core.numerics.basics import MathematicalProblem

__all__ = [
    "MathematicalProblem",
]




"""
Surrogate Model based on second order response surface equations.
"""

import numpy
from .base import SurrogateModel


class ResponseSurface(SurrogateModel):
    """
    Surrogate Model based on second order response surface equations.

    Attributes
    ----------
    betas : ndarray
        Vector of response surface equation coefficients.
    m : int
        Number of training points.
    n : int
        Number of independent variables.
    """

    def __init__(self):
        """
        Initialize all attributes.
        """
        self.m = 0  # number of training points
        self.n = 0  # number of independents
        # vector of response surface equation coefficients
        self.betas = numpy.zeros(0)

    def train(self, x, y):
        """
        Calculate response surface equation coefficients using least squares regression.

        Parameters
        ----------
        x : array-like
            Training input locations
        y : array-like
            Model responses at given inputs.
        """
        shape = numpy.shape(x)
        self.m = m = shape[0]
        self.n = n = shape[1]

        X = numpy.zeros((m, ((n + 1) * (n + 2)) // 2))

        # Modify X to include constant, squared terms and cross terms

        # Constant Terms
        X[:, 0] = 1.0

        # Linear Terms
        X[:, 1 : n + 1] = x

        # Quadratic Terms
        X_offset = X[:, n + 1 :]
        for i in range(n):
            # Z = numpy.einsum('i,ij->ij', X, Y) is equivalent to, but much faster and
            # memory efficient than, diag(X).dot(Y) for vector X and 2D array Y.
            # I.e. Z[i,j] = X[i]*Y[i,j]
            X_offset[:, : n - i] = numpy.einsum("i,ij->ij", x[:, i], x[:, i:])
            X_offset = X_offset[:, n - i :]

        # Determine response surface equation coefficients (betas) using least squares
        self.betas = numpy.linalg.lstsq(X, y, rcond=None)[0]

    def predict(self, x: numpy.ndarray) -> float:
        """
        Calculate predicted value of response based on the current response surface model.

        Parameters
        ----------
        x : array-like
            Point at which the surrogate is evaluated.

        Returns
        -------
        float
            Predicted response.
        """
        n = x.size
        X = numpy.zeros(((self.n + 1) * (self.n + 2)) // 2)

        # Modify X to include constant, squared terms and cross terms

        # Constant Terms
        X[0] = 1.0

        # Linear Terms
        X[1 : n + 1] = x

        # Quadratic Terms
        X_offset = X[n + 1 :]
        for i in range(n):
            X_offset[: n - i] = x[i] * x[i:]
            X_offset = X_offset[n - i :]

        # Predict new_y using X and betas
        return X.dot(self.betas)

    def linearize(self, x):
        """
        Calculate the jacobian of the Kriging surface at the requested point.

        Parameters
        ----------
        x : array-like
            Point at which the surrogate Jacobian is evaluated.

        Returns
        -------
        ndarray
            Jacobian of surrogate output wrt inputs.
        """
        n = self.n
        betas = self.betas

        x = x.flat

        jac = betas[1 : n + 1, :].copy()
        beta_offset = betas[n + 1 :, :]
        for i in range(n):
            jac[i, :] += x[i:].dot(beta_offset[: n - i, :])
            jac[i:, :] += x[i] * beta_offset[: n - i, :]
            beta_offset = beta_offset[n - i :, :]

        return jac.T



import os
import json

from typing import Any, Dict, Union

from collections import OrderedDict
import logging
import base64

from cosapp.systems import System
from cosapp.ports.port import BasePort, ExtensiblePort

from cosapp.utils.helpers import check_arg

logger = logging.getLogger(__name__)


def _get_tree_dict(
    obj: Union[System, BasePort],
    include_orphan_vars=True
) -> Dict[str, Any]:
    """Get a dictionary representation of the system hierarchy.
    """
    def init_tree(obj, subtype):
        return OrderedDict(
            name = obj.name,
            type = 'subsystem',
            subsystem_type = subtype,
        )

    def port_tree(port):
        tree_dict = init_tree(port, 'component')
        tree_dict['children'] = children = list()

        if include_orphan_vars or not isinstance(port, ExtensiblePort):
            dtype = 'param' if port.is_input else 'unknown'
            children.extend(
                map(lambda n: OrderedDict(name=n, type=dtype), iter(port))
            )
        return tree_dict

    def system_tree(system):
        tree_dict = init_tree(system, 'group')
        tree_dict['children'] = children = list()

        def add_ports(ports):
            nonlocal children
            children.extend(port_tree(port)
                for port in filter(lambda p: len(p) > 0, ports)
            )

        add_ports(system.inputs.values())

        children.extend(system_tree(child)
            for child in system.children.values()
        )
        # outputs come after child list for a matter of visualization
        add_ports(system.outputs.values())
        
        return tree_dict

    if isinstance(obj, System):
        return system_tree(obj)
    elif isinstance(obj, BasePort):
        return port_tree(obj)
    else:
        raise TypeError(f"Object must be a port or a system; got {type(obj)}.")


def _get_connections(system: System, include_orphan_vars: bool) -> Dict[str, str]:
    """Get a dictionary representation of the system connections. Structure is {'_in': '_out'}."""
    connections = dict()

    def contextual_name(port: BasePort):
        return port.name if port.owner is system else port.contextual_name

    def add_prefix(name: str, varname: str):
        prefix = f"{name}."
        return varname if varname.startswith(prefix) else f"{prefix}{varname}"

    # Gather connections for current system
    for connector in system.all_connectors():
        if not include_orphan_vars and isinstance(connector.sink, ExtensiblePort):
            continue
        sink_name = contextual_name(connector.sink)
        source_name = contextual_name(connector.source)
        for target, origin in connector.mapping.items():
            connections[f"{sink_name}.{target}"] = f"{source_name}.{origin}"

    # Recursively gather children's connections
    for name, child in system.children.items():
        child_connections = _get_connections(child, include_orphan_vars)
        for k, v in child_connections.items():
            connections[add_prefix(name, k)] = add_prefix(name, v)

    return connections


def _get_viewer_data(system: System, include_orphan_vars: bool) -> Dict[str, str]:
    """Get the data needed by the N2 viewer as a dictionary."""
    if isinstance(system, System):
        head = system
    else:
        raise TypeError('get_model_viewer_data only accepts System')

    if not include_orphan_vars:
        logger.warning('The system may contain inwards or outwards.')

    tree_data =_get_tree_dict(head, include_orphan_vars)
    connections = _get_connections(head, include_orphan_vars)
    data_dict = {
        'tree': tree_data,
        'connections_list' : [
            dict(src=origin, tgt=target)
            for target, origin in connections.items()
            if origin is not None
        ],
    }
    return data_dict


def view_model(
    problem_or_filename,
    outfile='n2.html',
    show_browser=True,
    embeddable=False,
    draw_potential_connections=False,
    include_orphan_vars=True,
) -> None:
    """
    Generates an HTML file containing a tree viewer. Optionally pops up a web browser to
    view the file.

    Parameters
    ----------
    problem_or_filename : A System
        System : The System for the desired tree.

    outfile : str, optional
        The name of the final output file

    show_browser : bool, optional
        If True, pop up the system default web browser to view the generated html file.
        Defaults to True.

    embeddable : bool, optional
        If True, gives a single HTML file that doesn't have the <html>, <DOCTYPE>, <body>
        and <head> tags. If False, gives a single, standalone HTML file for viewing.

    draw_potential_connections : bool, optional
        If true, allows connections to be drawn on the N2 that do not currently exist
        in the model. Defaults to True.

    include_orphan_vars : bool, optional
    If True, display inwards and outwards on the N2 diagram. Defaults to True.
    """
    def add_html_tags(content: str) -> str:
        """Format `content` into standalone HTML content."""
        html_begin_tags = """<!DOCTYPE html>
<html>
    <head>
        <meta http-equiv="Content-Type" content="text/html; charset=UTF-8">
    </head>
    <body>\n"""
        html_end_tags = """
    </body>
</html>"""
        return html_begin_tags + content + html_end_tags

    cur_dir = os.path.dirname(os.path.abspath(__file__))
    viz_dir = os.path.join(cur_dir, "visualization")
    lib_dir = os.path.join(viz_dir, "libs")
    src_dir = os.path.join(viz_dir, "src")
    sty_dir = os.path.join(viz_dir, "style")

    def get_content(path: str, filename: str, binary=False) -> str:
        fullpath = os.path.join(path, filename)
        if binary:
            with open(fullpath, "rb") as fp:
                return str(base64.b64encode(fp.read()).decode("ascii"))
        else:
            with open(fullpath, "r") as fp:
                return fp.read()

    def get_lib(filename: str, binary=False) -> str:
        return get_content(lib_dir, filename, binary)

    def get_src(filename: str, binary=False) -> str:
        return get_content(src_dir, filename, binary)

    def get_style(filename: str, binary=False) -> str:
        return get_content(sty_dir, filename, binary)

    # Get index.html
    index = get_content(viz_dir, "index.html")

    # Get model viewer data
    data = _get_viewer_data(problem_or_filename, include_orphan_vars)
    model_viewer_data = f"var modelData = {json.dumps(data)}"

    # Put all style and JS into index
    index = index.replace("{{awesomplete_style}}", get_style("awesomplete.css"))
    index = index.replace("{{partition_tree_style}}", get_style("partition_tree.css"))
    index = index.replace("{{fontello}}", get_style("fontello.woff", binary=True))
    index = index.replace("{{d3_lib}}", get_lib("d3.v4.min.js"))
    index = index.replace("{{awesomplete_lib}}", get_lib("awesomplete.js"))
    index = index.replace("{{vk_beautify_lib}}", get_lib("vkBeautify.js"))
    index = index.replace("{{model_data}}", model_viewer_data)
    index = index.replace("{{constants_lib}}", get_src("constants.js"))
    index = index.replace("{{modal_lib}}", get_src("modal.js"))
    index = index.replace("{{svg_lib}}", get_src("svg.js"))
    index = index.replace("{{search_lib}}", get_src("search.js"))
    index = index.replace("{{legend_lib}}", get_src("legend.js"))
    index = index.replace("{{draw_lib}}", get_src("draw.js"))
    index = index.replace("{{ptn2_lib}}", get_src("ptN2.js"))
    index = index.replace(
        "{{draw_potential_connections}}",
        'true' if draw_potential_connections else 'false',
    )

    if not embeddable:
        index = add_html_tags(index)

    with open(outfile, 'w') as fp:
        fp.write(index)

    # Open in browser, if required
    if show_browser:
        from cosapp.tools.problem_viewer.webview import webview
        webview(outfile)



import pytest
from typing import Tuple

from cosapp.multimode.discreteStepper import DiscreteStepper
from cosapp.drivers import EulerExplicit
from cosapp.systems import System


class AcceleratedSystem(System):
    def setup(self):
        self.add_inward('a', 1.0)
        self.add_transient('v', der='a')
        self.add_transient('x', der='v')

        self.add_event('fly')
        self.add_event('stop')


@pytest.fixture
def accelerated_case() -> Tuple[System, DiscreteStepper]:
    system = AcceleratedSystem('system')
    driver = system.add_driver(
        EulerExplicit(
            'time_driver',
            time_interval=(0, 10)
        )  # arbitrary
    )
    system.a = 10
    system.fly.trigger = "v > 12.34"
    system.stop.trigger = "x > 2.58"
    handler = DiscreteStepper(driver)
    # Use exact solution as interpolators
    handler.sysview.interp = {
        'v': lambda t: system.a * t,
        'x': lambda t: 0.5 * system.a * t**2,
    }
    return system, handler


def test_DiscreteStepper_trigger_time(accelerated_case):
    system, stepper = accelerated_case
    stepper.interval = (0.5, 1.5)
    tf = stepper.trigger_time(system.fly)
    ts = stepper.trigger_time(system.stop)
    assert tf == pytest.approx(1.234)
    assert ts == pytest.approx(0.7183314)


def test_DiscreteStepper_event_detected(accelerated_case):
    system, stepper = accelerated_case
    stepper.interval = (0.5, 1.5)
    stepper.sysview.exec(0.5)
    stepper.shift()
    # Hack: manual event stepping just for testing
    for event in system.all_events():
        event.step()
    stepper.sysview.exec(0.7)
    stepper.shift()
    assert not stepper.event_detected()
    # Hack: manual event stepping just for testing
    for event in system.all_events():
        event.step()
    stepper.sysview.exec(1.5)
    stepper.shift()
    assert stepper.event_detected()


def test_DiscreteStepper_primal_event(accelerated_case):
    system, stepper = accelerated_case
    stepper.interval = (0.5, 1.5)
    stepper.sysview.exec(0.5)
    # Manual reevaluation and shifting for testing purposes
    stepper.reevaluate_primitive_events()
    stepper.shift()
    stepper.sysview.exec(1.5)  
    # Neither `event_detected()` nor `find_primal_event()` call the `step()` method of events
    assert stepper.event_detected()
    assert system.x > 2.58
    assert system.v > 12.34
    # Manual stepping for testing purposes
    for event in system.all_events():
        event.step()
    first = stepper.find_primal_event()
    assert first.event is system.stop
    assert first.time == pytest.approx(0.7183314)



"""
Classes connecting `Port` of foreign `System` to transfer variable values.
"""
import abc
import copy
import logging
import weakref
from types import MappingProxyType
from typing import Callable, Dict, Iterable, Iterator, List, Optional, Tuple, Union, Any, Type

from cosapp.ports import units
from cosapp.ports.port import BasePort, Port
from cosapp.utils.helpers import check_arg, is_numerical

logger = logging.getLogger(__name__)


class ConnectorError(Exception):
    """Raised if a connector cannot be created between two `Port`.

    Attributes
    ----------
    message : str
        Error message
    """

    def __init__(self, message: str):
        """Instantiate a error object from the error descriptive message.

        Parameters
        ----------
        message : str
            Error message
        """
        self.message = message


class BaseConnector(abc.ABC):
    """This class connect two ports without enforcing that all port variables are connected.

    The link is oriented from the source to the sink.

    Parameters
    ----------
    name : str
        Name of the connector
    source : BasePort
        Port from which originate the variables
    mapping : str or List[str] or Dict[str, str]
        (List of) common name(s) or mapping name dictionary
    sink : BasePort
        Port to which the variables are transferred
    """

    def __init__(
        self,
        name: str,
        sink: BasePort,
        source: BasePort,
        mapping: Union[str, List[str], Dict[str, str], None] = None,
    ):
        """Connector constructor from the two `BasePort` to link and the list of variables to map.

        If no mapping is provided, connection will be made between variables based on their names. 
        If a name mapping is provided as a list, the name should be present in both port. And if 
        the mapping is specified as a dictionary, the keys belong to the sink port and the values
        to the source port.

        Parameters
        ----------
        name : str
            Name of the connector
        sink : BasePort
            Port to which the variables are transferred.
        source : BasePort
            Port from which originate the variables.
        mapping : str or List[str] or Dict[str, str], optional
            (List of) common name(s) or mapping name dictionary; default None (i.e. no mapping).
        
        Raises
        ------
        ConnectorError
            If the connection between the `source` and the `sink` is not possible to establish.
        """
        self.__check_port(sink, 'sink')
        self.__check_port(source, 'source')
        check_arg(name, 'name', str, lambda s: len(s.strip()) > 0)

        if source is sink:
            raise ConnectorError("Source and sink cannot be the same object.")

        # Generate mapping dictionary
        if mapping is None:
            mapping = dict((v, v) for v in sink if v in source)
        else:
            mapping = self.format_mapping(mapping)

        self._name = name  # type: str
        self._mapping = mapping  # type: Dict[str, str]
        self._source = self.__get_port(source, sink=False, check=False)  # type: weakref.ReferenceType[BasePort]
        self._sink = self.__get_port(sink, sink=True, check=False)  # type: weakref.ReferenceType[BasePort]

    @abc.abstractmethod
    def transfer(self) -> None:
        """Transfer values from `source` to `sink`."""
        pass

    def __repr__(self) -> str:
        mapping = self._mapping
        if self.preserves_names():
            mapping = list(mapping)
        return "{}({} <- {}, {})".format(
            type(self).__qualname__,
            self.sink.contextual_name,
            self.source.contextual_name,
            mapping,
        )

    @property
    def name(self) -> str:
        """str : name of the connector."""
        return self._name

    @property
    def source(self) -> BasePort:
        """`BasePort`: Port from which transferred values originate."""
        return self._source()

    @source.setter
    def source(self, port: BasePort) -> None:
        self._source = self.__get_port(port, sink=False, check=True)

    @property
    def sink(self) -> BasePort:
        """`BasePort`: Port to which values are transferred."""
        return self._sink()

    @sink.setter
    def sink(self, port: BasePort) -> None:
        self._sink = self.__get_port(port, sink=True, check=True)

    def __get_port(self, port: BasePort, sink: bool, check=True) -> "weakref.ref[BasePort]":
        """Returns a weakref to `port`, after compatibility check with internal mapping."""
        port_type = 'sink' if sink else 'source'

        if check:
            self.__check_port(port, port_type)

        varnames: Iterator[str] = getattr(self, f"{port_type}_variables")()
        not_found = list(
            filter(lambda name: name not in port, varnames)
        )
        if not_found:
            if len(not_found) == 1:
                message = f"variable {not_found[0]!r} does not exist"
            else:
                message = f"variables {not_found} do not exist"
            raise ConnectorError(
                f"{port_type.title()} {message} in port {port}."
            )
        
        return weakref.ref(port)

    @staticmethod
    def __check_port(port: BasePort, name: str) -> None:
        check_arg(port, name, BasePort, stack_shift=1)

        if not port.owner:
            raise ConnectorError(f"{name.title()} owner is undefined.")

    def __len__(self) -> int:
        return len(self._mapping)

    @property
    def mapping(self) -> Dict[str, str]:
        """Dict[str, str] : Variable name mapping between the sink (key) and the source (value)."""
        return MappingProxyType(self._mapping)

    def sink_variables(self) -> Iterator[str]:
        return self._mapping.keys()

    def source_variables(self) -> Iterator[str]:
        return self._mapping.values()

    def sink_variable(self, source_variable: str) -> str:
        """Returns the name of the sink variable associated to `source_variable`"""
        for sink, source in self._mapping.items():
            if source == source_variable:
                return sink
        raise KeyError(source_variable)

    def source_variable(self, sink_variable: str) -> str:
        """Returns the name of the source variable associated to `sink_variable`"""
        return self._mapping[sink_variable]

    def preserves_names(self) -> bool:
        """Returns `True` if connector mapping preserves variable names,
        `False` otherwise."""
        return all(target == origin for target, origin in self._mapping.items())

    def is_mirror(self) -> bool:
        """Returns `True` if connector is an identical, one-to-one mapping
        between two ports of the same kind; `False` otherwise."""
        sink, source = self.sink, self.source
        return (
            type(sink) is type(source)
            and isinstance(sink, Port)
            and len(self) == len(sink)
            and self.preserves_names()
        )

    def update_mapping(self, mapping: Dict[str, str]) -> None:
        """Extend current mapping with additional dictionary.

        Parameters
        ----------
        mapping : Dict[str, str]
            Variable name mapping extending current mapping.
        """
        self._mapping.update(mapping)

    def remove_variables(self, names: Iterable[str]) -> None:
        """Remove the provided variables from this connection.

        The provided names should be sink names.

        Parameters
        ----------
        names : Iterable[str]
            Collection of sink variable names to be removed.
        """
        for variable in names:
            del self._mapping[variable]

    def info(self) -> Union[Tuple[str, str], Tuple[str, str, Dict[str, str]]]:
        """Returns connector information in a tuple.

        If the name mapping is complete, with identical names,
        it is omitted, and the output tuple is formatted as:
        - (target_name, source_name)

        Otherwise, output is formatted as:
        - (target_name, source_name, name_mapping)

        Returns
        -------
        tuple
            Tuple representing connector
        """
        # If the mapping is full and with the same nomenclature
        target, origin = self.port_names()
        same_nomenclature = False
        if len(self._mapping) == len(self.source) == len(self.sink):
            same_nomenclature = self.preserves_names()
        if same_nomenclature:
            info = (target, origin)
        else:
            info = (target, origin, self._mapping.copy())
        return info

    def to_dict(self) -> Dict[str, Union[Tuple[str, str], Tuple[str, str, Dict[str, str]]]]:
        """Converts connector into a single-key dictionary.
        The key is the connector name; associated value
        is the tuple returned by method `info()`.

        Returns
        -------
        dict
            Dictionary {name: info_tuple} representing connector
        """
        return {self.name: self.info()}

    def port_names(self) -> Tuple[str, str]:
        """Returns source and sink contextual names as a str tuple.

        Returns
        -------
        tuple
            (source_name, sink_name) tuple
        """
        source, sink = self.source, self.sink

        if source.owner is sink.owner.parent:
            origin = source.name
        else:
            origin = source.contextual_name

        if sink.owner is source.owner.parent:
            target = sink.name
        else:
            target = sink.contextual_name
        
        return target, origin

    @staticmethod
    def format_mapping(mapping: Union[str, List[str], Dict[str, str]]) -> Dict[str, str]:
        """Returns suitable name mapping for connectors,
        from different kinds of argument `mapping`.

        Parameters:
        -----------
        - mapping, Union[str, List[str], Dict[str, str]]:
            Name mapping, given as either a string (single variable),
            a list of strings, or a full name mapping, as a dictionary.
        
        Returns:
        --------
        Dict[str, str]: suitable mapping for connectors.
        """
        if isinstance(mapping, str):
            mapping = {mapping: mapping}
        elif isinstance(mapping, MappingProxyType):
            mapping = dict(mapping)
        elif not isinstance(mapping, dict):
            mapping = dict(zip(mapping, mapping))
        return mapping


class Connector(BaseConnector):
    """Shallow copy connector.
    See `BaseConnector` for base class details.
    """
    def __init__(
        self,
        name: str,
        sink: BasePort,
        source: BasePort,
        mapping: Union[str, List[str], Dict[str, str], None] = None,
    ):
        super().__init__(name, sink, source, mapping)

        self._unit_conversions = {} # type: Dict[str, Optional[Tuple[float, float]]]
        self._transfer_func = {}  # type: Dict[str, Callable[[Any], Any]]
        self.update_unit_conversion()

    @BaseConnector.source.setter
    def source(self, port: BasePort) -> None:
        cls = self.__class__
        super(cls, cls).source.__set__(self, port)
        self.update_unit_conversion()

    @BaseConnector.sink.setter
    def sink(self, port: BasePort) -> None:
        cls = self.__class__
        super(cls, cls).sink.__set__(self, port)
        self.update_unit_conversion()

    def update_mapping(self, mapping: Dict[str, str]) -> None:
        super().update_mapping(mapping)
        self.update_unit_conversion()

    def remove_variables(self, names: Iterable[str]) -> None:
        super().remove_variables(names)
        for variable in names:
            del self._unit_conversions[variable]
            del self._transfer_func[variable]

    def update_unit_conversion(self, name: Optional[str] = None) -> None:
        """Update the physical unit conversion on the connector.

        If `name` is not `None`, update the conversion only for the connexion towards that variable.

        Parameters
        ----------
        name : str, optional
            Name of the variable for which unit conversion needs an update; default None (i.e. all
            conversions will be updated).

        Raises
        ------
        UnitError
            If unit conversion from source to sink is not possible
        """
        source, sink = self.source, self.sink
        mapping = self.mapping

        def update_one_connection(key: str) -> None:
            """Update the unit converter of the connected key.

            Parameters
            ----------
            key : str
                Name of the connected variable for which the unit converter should be updated.
            """
            target = sink.get_details(key)
            origin = source.get_details(mapping[key])
            has_unit = {
                'target': bool(target.unit),
                'origin': bool(origin.unit),
            }
            if has_unit['origin'] != has_unit['target']:
                # Send a warning if one end of the connector
                # has a physical unit and the other is dimensionless
                message = lambda origin_status, target_status: (
                    f"Connector source {origin.full_name!r} {origin_status}, but target {target.full_name!r} {target_status}."
                )
                if has_unit['origin']:
                    logger.warning(
                        message(f"has physical unit {origin.unit}", "is dimensionless")
                    )
                else:
                    logger.warning(
                        message("is dimensionless", f"has physical unit {target.unit}")
                    )
            # Get conversion constants between units
            constants = units.get_conversion(origin.unit, target.unit)
            if constants is None and is_numerical(sink[key]):
                constants = (1.0, 0.0)  # legacy - tests be must changed if suppressed
            self._unit_conversions[key] = constants

            # Get transfer function
            try:
                factor, offset = self._unit_conversions[key]
            except:
                pass
            else:
                if (factor, offset) != (1, 0):
                    self._transfer_func[key] = lambda value: factor * (value + offset)

        if name is None:
            for name in mapping:
                update_one_connection(name)
        else:
            update_one_connection(name)

    def transfer(self) -> None:
        source, sink = self.source, self.sink

        for target, origin in self._mapping.items():
            # get/setattr faster for Port
            value = getattr(source, origin)
            transfer = self._transfer_func.get(target, copy.copy)
            setattr(sink, target, transfer(value))


def MakeDirectConnector(classname: str, transform: Optional[Callable]=None, **kwargs) -> Type[BaseConnector]:
    """Connector factory using a simple transfer function, with no unit conversion.
    """
    if transform is None:
        transfer_attr = lambda sink, name, value: setattr(sink, name, value)
    else:
        transfer_attr = lambda sink, name, value: setattr(sink, name, transform(value))

    class DirectConnector(BaseConnector):
        """Connector with simple transfer function"""
        def transfer(self) -> None:
            source, sink = self.source, self.sink

            for target, origin in self._mapping.items():
                value = getattr(source, origin)
                transfer_attr(sink, target, value)

    return type(classname, (DirectConnector,), kwargs)


PlainConnector = MakeDirectConnector(
    "PlainConnector",
    __doc__ = (
        "Plain assignment connector, with no unit conversion."
        " Warning: may generate common references between sink and source variables."
    ),
)

CopyConnector = MakeDirectConnector(
    "CopyConnector", copy.copy,
    __doc__ = "Shallow copy connector, with no unit conversion.",
)

DeepCopyConnector = MakeDirectConnector(
    "DeepCopyConnector", copy.deepcopy,
    __doc__ = "Deep copy connector, with no unit conversion.",
)



from cosapp.base import System
import numpy as np


class PointDynamics1D(System):
    """Point mass dynamics in 1D using scalars"""
    def setup(self):
        self.add_inward("mass", 1.0, unit='kg')
        self.add_inward("acc_ext", 0.0, unit='m/s**2')
        self.add_inward("force_ext", 0.0, unit='N')

        self.add_outward("force", 0.0, unit='m/s**2')
        self.add_outward("acc", 0.0, unit='N')

    def compute(self):
        self.force = self.force_ext + self.mass * self.acc_ext
        self.acc = self.force / self.mass


class PointDynamics(System):
    """Point mass dynamics"""
    def setup(self):
        self.add_inward("mass", 1.0, unit='kg')
        self.add_inward("acc_ext", unit='m/s**2')
        self.add_inward("force_ext", unit='N')

        self.add_outward("force", unit='m/s**2')
        self.add_outward("acc", unit='N')

    def compute(self):
        self.force = self.force_ext + self.mass * self.acc_ext
        self.acc = self.force / self.mass


class PointFriction(System):
    """Point mass ~ v^2 friction model"""
    def setup(self):
        self.add_inward('v', desc="Velocity")
        self.add_inward('cf', 0.1, desc="Friction coefficient")

        self.add_outward("force", unit='N')

    def compute(self):
        self.force = (-self.cf * np.linalg.norm(self.v)) * self.v


class PointMass(System):
    def setup(self, g=[0, 0, -9.81]):
        self.add_child(PointFriction('friction'), pulling=['cf', 'v'])
        self.add_child(PointDynamics('dynamics'), pulling={
            'mass': 'mass',
            'force': 'force',
            'acc_ext': 'g',
            'acc': 'a',
        })
        self.exec_order = ['friction', 'dynamics']

        self.connect(self.friction, self.dynamics, {"force": "force_ext"})

        self.g = np.array(g)
        self.v = np.zeros_like(g)
        self.run_once()  # propagate values to initialize array dimensions

        self.add_transient('v', der='a')
        self.add_transient('x', der='v')


class BouncingBall(System):
    """Bouncing point mass"""
    def setup(self, g=[0, 0, -9.81]):
        self.add_child(PointMass('point', g=g), pulling=[
            'x', 'v', 'a', 'mass', 'cf', 'g',
        ])
        self.add_event('rebound', trigger="x[-1] <= 0")
        self.add_inward('cr', 1.0, desc="Rebound restitution coefficient", limits=(0, 1))
        self.add_outward_modevar("n_rebounds", init=0, dtype=int)

    def transition(self):
        cr = max(0, min(self.cr, 1))
        if self.rebound.present:
            self.n_rebounds += 1
            v = self.v
            if abs(v[-1]) < 1e-6:
                v[-1] = 0
            else:
                v[-1] *= -cr



"""
Basic class handling model tree structure.
"""
from __future__ import annotations
import abc
import collections
import logging
from numbers import Integral
from typing import (
    Optional, Any, List, Dict, OrderedDict,
    Generator, MappingView, Sequence,
    TypeVar,
)

from cosapp.patterns.visitor import Visitor, Component as VisitedComponent
from cosapp.core.signal import Signal
from cosapp.utils.naming import NameChecker, CommonPorts
from cosapp.utils.helpers import check_arg
from cosapp.utils.logging import LoggerContext, LogFormat, LogLevel

logger = logging.getLogger(__name__)

Child = TypeVar("Child", bound="Module")


class Module(LoggerContext, VisitedComponent, metaclass=abc.ABCMeta):
    # class Module does not inherit directly from abc.ABC, due to 
    # a bug in Python 3.6 preventing the use of __weakref__ in slots.
    # Ref: https://bugs.python.org/issue30463
    # Here, use of abc.ABCMeta makes the class effectively abstract,
    # by activating decorators of the kind @abc.abstractclass, etc.
    """
    A class to describe generic properties and functions of a component that can be single or
    made of child `Module`.

    Parameters
    ----------
    - name [str]:
        `Module` name

    Attributes
    ----------
    - name [str]:
        `Module` name
    - children [dict[str, Module]]:
        Sub-modules of current `Module`, referenced by names.
    - parent [Module]:
        Parent `Module` of current `Module`; `None` if there is no parent.
    - exec_order [Iterator[str]]:
        Execution order in which sub-modules should be computed.
    - description [str]:
        `Module` description.

    - _active [bool]:
        If False, the `Module` will not execute its `run_once` method
    - _compute_calls [int]:
        Store if the number of times :py:meth:`~cosapp.core.module.Module.compute` was called (due to inhibition of clean status)

    Signals
    -------
    - setup_ran [Signal]:
        Signal emitted after :py:meth:`~cosapp.core.module.Module.call_setup_run` execution
    - computed [Signal]:
        Signal emitted after the :py:meth:`~cosapp.core.module.Module.compute` stack (= after :py:meth:`~cosapp.core.module.Module._post_compute`) execution
    - clean_ran [Signal]:
        Signal emitted after the :py:meth:`~cosapp.core.module.Module.call_clean_run` execution
    """

    __slots__ = (
        '__weakref__', '_name', 'children', 'parent', '_active',
        '_compute_calls', 'setup_ran', 'computed', 'clean_ran',
        '__members', '_desc',
    )

    _name_check = NameChecker(excluded=CommonPorts.names())

    def __init__(self, name: str):
        """`Module` constructor

        Parameters
        ----------
        - name [str]:
            Module name
        """
        self._name = self._name_check(name)
        self._desc = ""
        self.children: Dict[str, Module] = collections.OrderedDict()
        self.parent: Optional[Module] = None
        self._active: bool = True
        self._compute_calls: int = 0

        # Signals
        self.setup_ran = Signal(name="cosapp.core.module.Module.setup_ran")
        self.clean_ran = Signal(name="cosapp.core.module.Module.clean_ran")
        self.computed = Signal(name="cosapp.core.module.Module.computed")
        private_prefix = f"_{type(self).__qualname__}__"
        self.__members = set(
            filter(
                lambda name: not name.startswith(private_prefix),
                dir(type(self))
            )
        )

    def __dir__(self):
        """Collection of all member names (used for autocompletion)"""
        return self.__members

    def tree(self, downwards=False) -> Generator[Module, None, None]:
        """Generator recursively yielding all elements in module tree.
        
        Parameters:
        -----------
        - downwards [bool, optional]:
            If `True`, yields elements from top to bottom.
            If `False` (default), yields elements from bottom to top.
        """
        if downwards:
            yield self
        for child in self.children.values():
            yield from child.tree(downwards)
        if not downwards:
            yield self

    def send_visitor(self, visitor: Visitor, downwards=False) -> None:
        """Recursively accept visitor throughout module tree."""
        for module in self.tree(downwards):
            module.accept(visitor)

    def __getattr__(self, name: str) -> Any:
        try:  # Faster than testing
            return self.children[name]
        except KeyError:
            return super().__getattribute__(name)

    @property
    def contextual_name(self) -> str:
        """str : Name of the module relative to the root one."""
        return "" if self.parent is None else self.full_name(trim_root=True)

    @property
    def compute_calls(self) -> int:
        """int : Number of calls to the compute method at last execution."""
        return self._compute_calls

    @property
    def name(self) -> str:
        """str : `Module` identifier."""
        return self._name

    @name.setter
    def name(self, name: str) -> None:
        self._name = self._name_check(name)

    @property
    def description(self) -> str:
        """str: Module description"""
        return self._desc

    @description.setter
    def description(self, desc: str) -> None:
        check_arg(desc, 'description', str)
        self._desc = desc

    @property
    def exec_order(self) -> MappingView[str]:
        """MappingView[str]: sub-module execution order, as a name iterator"""
        return self.children.keys()

    @exec_order.setter
    def exec_order(self, namelist: Sequence[str]) -> None:
        if not isinstance(namelist, Sequence):
            raise TypeError("exec_order must be an ordered sequence of strings")
        nameset = set(self.children)
        if set(namelist) != nameset:
            if nameset:
                msg = f"exec_order must be a permutation of {list(self.children)}"
            else:
                msg = f"Can't set exec_order, as {self.name!r} has no children"
            logger.error(f"{msg}; got {namelist}.")
            raise ValueError(msg)
        elif len(namelist) > len(self.children):
            repeated = list(namelist)
            for name in set(namelist):
                repeated.remove(name)
            raise ValueError(f"Repeated items {sorted(set(repeated))}")
        # Rearrange children in a new dictionary
        self.children = OrderedDict(
            (name, self.children[name])
            for name in namelist
        )

    @property
    def size(self) -> int:
        """int: Total number of elements in tree."""
        return sum(1 for _ in self.tree())

    def _add_member(self, name: str) -> None:
        """Add `name` to dynamic member list"""
        if '.' not in name:
            self.__members.add(name)

    def _pop_member(self, name: str) -> None:
        """Remove `name` to dynamic member list"""
        try:
            self.__members.remove(name)
        except:
            pass

    def path_to_root(self) -> Generator[Module, None, None]:
        """Generator recursively yielding all elements up to root module.
        """
        current = self
        yield current
        while current.parent is not None:
            current = current.parent
            yield current

    def root(self) -> Module:
        for root in self.path_to_root():
            continue
        return root

    def path(self) -> List[Module]:
        """Returns full path from root Module as a list.
        
        Returns
        -------
        List[Module]
            Full module list from root to self
        """
        path = list(self.path_to_root())
        return list(reversed(path))

    def path_namelist(self) -> List[str]:
        """Returns full name list from root Module.
        
        Returns
        -------
        List[str]
            The module full name list
        """
        names = [elem.name for elem in self.path_to_root()]
        return list(reversed(names))

    def full_name(self, trim_root=False) -> str:
        """Returns full name from root Module.
        
        Parameters
        ----------
        trim_root : bool (optional, default False)
            Exclude root Module name if True.

        Returns
        -------
        str
            The module full name
        """
        names = self.path_namelist()
        start = 1 if trim_root else 0
        return ".".join(names[start:])

    def call_setup_run(self):
        """Execute `setup_run` recursively on all modules."""
        with self.log_context(" - call_setup_run"):
            logger.debug(f"Call {self.name}.setup_run")
            self._compute_calls = 0  # Reset the counter
            for child in self.children.values():
                child.call_setup_run()
            self.setup_run()
            self.setup_ran.emit()

    def setup_run(self):
        """Method called once before starting any simulation."""
        pass  # pragma: no cover

    def call_clean_run(self):
        """Execute `clean_run` recursively on all modules."""
        with self.log_context(" - call_clean_run"):
            logger.debug(f"Call {self.name}.clean_run")
            self.clean_run()
            for child in self.children.values():
                child.call_clean_run()
            self.clean_ran.emit()

    def clean_run(self):
        """Method called once after any simulation."""
        pass  # pragma: no cover

    def get_path_to_child(self, other: Module) -> str:
        """
        Returns the relative path to target Module `other`.
        Raises `ValueError` if `other` is not related to current Module.

        Returns
        -------
        Roots
            The relative path to `target`
        """
        path = list()
        child = other
        while child is not self:
            path.append(child.name)
            child = child.parent
            if child is None:
                raise ValueError(
                    f"{other.name!r} is not a child of {self.name!r}."
                )
        return ".".join(reversed(path))

    def add_child(self, child: Child, execution_index: Optional[int]=None, desc="") -> Child:
        """Add a child `Module` to the current `Module`.

        When adding a child `Module`, it is possible to specified its position
        in the execution order.

        Parameters
        ----------
        - child [Module]:
            `Module` to add to the current `Module`
        - execution_index [int, optional]:
            Index of the execution order list at which the `Module` should be inserted;
            default latest.
        - desc [str, optional]:
            Module description in the context of its parent module.

        Returns
        -------
        `child`
        """
        # Type validation
        check_arg(child, 'child', Module)

        specific_order = None
        if execution_index is not None:
            check_arg(execution_index, 'execution_index', Integral, lambda i: i >= 0)
            specific_order = list(self.exec_order)
            specific_order.insert(execution_index, child.name)

        if child.name in self.children:
            raise ValueError(
                "{} {!r} cannot be added, as Module already contains an object with the same name"
                "".format(type(child).__qualname__, child.name)
            )

        child.parent = self
        child.description = desc
        self.children[child.name] = child
        self._add_member(child.name)

        if specific_order:
            self.exec_order = specific_order

        return child

    def pop_child(self, name: str) -> Module:
        """Remove submodule `name` from current module.

        Parameters
        ----------
        name: str
            Name of submodule to be removed.

        Returns
        -------
        `Module`
            The removed module.

        Raises
        ------
        `AttributeError` if no match is found.
        """
        try:
            child = self.children.pop(name)
        except KeyError:
            raise AttributeError(f"Component {name} is not a child of {self}.")

        self._pop_member(name)
        child.parent = None

        return child

    def _precompute(self) -> None:
        """Actions performed prior to the `Module.compute` call."""
        pass

    def compute_before(self) -> None:
        """Contains the customized `Module` calculation, to execute before children."""
        pass

    def compute(self) -> None:
        """Contains the customized `Module` calculation, to execute after children."""
        pass

    def _postcompute(self) -> None:
        """Actions performed after the `Module.compute` call."""
        pass

    def run_once(self) -> None:
        """Run the module once.

        Execute the model of this `Module` and its children in the execution order.

        Notes
        -----

        The driver are not executed when calling this method; only the physical model.
        """
        with self.log_context(" - run_once"):
            if self.is_active():
                self._precompute()

                logger.debug(f"Call {self.name}.compute_before()")
                self.compute_before()

                for name, child in self.children.items():
                    logger.debug(f"Call {self.name}.{name}.run_once()")
                    child.run_once()

                logger.debug(f"Call {self.name}.compute()")
                self._compute_calls += 1
                self.compute()

                self._postcompute()
                self.computed.emit()
            
            else:
                logger.debug(f"Skip {self.name} execution - Inactive")

    def is_active(self) -> bool:
        """Is this Module execution activated?

        Returns
        -------
        bool
            Activation status
        """
        return self._active

    @abc.abstractmethod
    def is_standalone(self) -> bool:
        """Is this Module able to solve itself?

        Returns
        -------
        bool
            Ability to solve the module or not.
        """
        pass

    def log_debug_message(
        self,
        handler: "HandlerWithContextFilters",
        record: logging.LogRecord,
        format: LogFormat = LogFormat.RAW,
    ) -> bool:
        """Callback method on the module to log more detailed information.
        
        This method will be called by the log handler when :py:meth:`~cosapp.utils.logging.LoggerContext.log_context`
        is active if the logging level is lower or equals to VERBOSE_LEVEL. It allows
        the object to send additional log message to help debugging a simulation.

        .. note::
            logger.log method cannot be used here. Use handler.handle(record)

        Parameters
        ----------
        handler : HandlerWithContextFilters
            Log handler on which additional message should be published.
        record : logging.LogRecord
            Log record
        format : LogFormat
            Format of the message

        Returns
        -------
        bool
            Should the provided record be logged?
        """
        message = record.getMessage()
        emit_record = super().log_debug_message(handler, record, format)

        if message.endswith("call_setup_run"):
            emit_record = False
        elif message.endswith("call_clean_run"):
            emit_record = False
            activate = getattr(record, "activate", None)
            if activate == False:
                # Display the number of system execution
                msg = f"Compute calls for {self.full_name()}: {self.compute_calls}"
                handler.log(LogLevel.DEBUG, msg, name=logger.name)

        return emit_record



from __future__ import annotations
from abc import abstractmethod
from numbers import Number
from typing import Any, Dict, Optional, Tuple, FrozenSet, Union
from collections.abc import Collection

import numpy

from cosapp.core.eval_str import EvalString
from cosapp.utils.helpers import check_arg
from cosapp.utils.naming import natural_varname


class AbstractResidue:
    """Abstract class defining residue for solvers.

    Parameters
    ----------
    - context : cosapp.core.module.Module
        CoSApp Module in which this residue is defined
    - name : str
        Residue name
    """
    def __init__(
        self,
        context: "System",
        name: str,
    ):
        from cosapp.systems import System
        check_arg(context, "context", System)
        check_arg(name, "name", str)
        self._reference_value = 1.0
        self._value = None  # type: Union[Number, numpy.ndarray]
        self._context = context  # type: System
        self._name = natural_varname(name)  # type: str
        super().__init__()

    def __str__(self) -> str:
        return f"{self._name} := {self._value}"

    def __repr__(self) -> str:
        return f"{type(self).__qualname__}({self._name}): {self._value!s}"

    def _get_zeros_rhs(self, lhs: Union[str, Number, numpy.ndarray]) -> Union[str, Number, numpy.ndarray]:
        """Generate the default zeroed right-hand side of the residue equation.

        The default right-hand side is a zero value or vector. If the left
        hand side is defined by a string to be evaluated, the right hand
        side will be expressed as a string to evaluate.

        Attributes
        ----------
        lhs : str
            Left-hand side of the residue equation

        Returns
        -------
        Union[str, Number, numpy.ndarray]
            Default zeroed right-hand side
        """
        if isinstance(lhs, str):
            lhs_evaluated = EvalString(lhs, self.context).eval()
            return f"zeros({numpy.shape(lhs_evaluated)})"
        else:
            return numpy.zeros_like(numpy.asarray(lhs))

    @abstractmethod
    def update(self, **kwargs) -> Union[Number, numpy.ndarray]:
        """Update the value of the residue.

        Returns
        -------
        Number or numpy.ndarray
            The updated value
        """
        # TODO kwargs is bad
        pass

    @property
    def reference(self) -> Union[Number, numpy.ndarray]:
        """Number or numpy.ndarray: Reference value to normalize the residue with."""
        return self._reference_value

    @reference.setter
    def reference(self, value: Union[Number, numpy.ndarray]) -> None:
        """Set the reference value.

        Parameters
        ----------
        value : Union[Number, numpy.ndarray]
            User defined reference value; default None (guessed by the solver)
        """
        self._reference_value = value

    @property
    def context(self) -> "System":
        """System: CoSApp system usable in string evaluation."""
        return self._context

    @property
    def name(self) -> str:
        """str : Residue name"""
        return self._name

    @property
    def value(self) -> Union[Number, numpy.ndarray]:
        """Number or numpy.ndarray: residue's value."""
        return self._value

    @abstractmethod
    def copy(self) -> AbstractResidue:
        """Copy the object.

        Returns
        -------
        AbstractResidue
            The duplicated object.
        """
        pass


class Residue(AbstractResidue):
    """Classical residue definition based on an equality equation

    Left-hand side == Right-hand side

    The right-hand side is assumed null if not specified.

    The residue can be defined by providing directly the numerical values
    of the left and right-hand sides. An alternative is to provide an string
    expression to be evaluated.

    Parameters
    ----------
    - context : cosapp.systems.System
        CoSApp System to which residue is linked
    - lhs : Number or numpy.ndarray or str
        Left-hand side of the equation.
    - rhs : Number or numpy.ndarray or str or None, optional
        Right-hand side of the equation; default None (i.e. equals zeros).
    - name : str or None, optional
        Residue name; default None (built from lhs and rhs)
    - reference: Number or numpy.ndarray or None
        Reference value to normalized the residue; default is unity
    """

    @staticmethod
    def residue_norm(
        left: Union[Number, Collection, numpy.ndarray],
        right: Union[Number, Collection, numpy.ndarray] = None,
    ) -> Union[Number, numpy.ndarray]:
        """Computes the order of magnitude of left- and right-hand sides to approximate the reference value.

        Parameters
        ----------
        - left : float or list of float or numpy.ndarray
            Left-hand side of the equation
        - right : float or list of float or numpy.ndarray or None (default)
            Right-hand side of the equation if not None

        Returns
        -------
        Number or numpy.ndarray
            An approximated reference value
        """
        a = numpy.abs(numpy.asarray(left, float))
        if right is not None:
            a += numpy.abs(right)
        return 10 ** numpy.floor(numpy.log10(numpy.where(a == 0, 1, a)))

    @staticmethod
    def evaluate_residue(
        lhs: Union[Number, Collection, numpy.ndarray],
        rhs: Union[Number, Collection, numpy.ndarray],
        reference: Union[Number, Collection, numpy.ndarray, str] = 1,
    ) -> Union[float, numpy.ndarray]:
        """Evaluate the normalized residue from left- and right-hand sides of an equation.

        The formula depends on the left- and right-hand side value of the equation:
            residues = (lhs - rhs) / reference

        Parameters
        ----------
        - lhs : Number or list of float or numpy.ndarray
            Left-hand side of the equation
        - rhs : Number or list of float or numpy.ndarray
            Right-hand side of the equation
        - reference : Number or list of float or numpy.ndarray or "norm"
            Reference value to normalize the equation with; default is unity

        Returns
        -------
        Number or numpy.ndarray
            Normalized residue
        """
        if isinstance(reference, str) and reference == "norm":  # type check avoids a warning for numpy arrays
            reference = Residue.residue_norm(lhs, rhs)
        return (numpy.asarray(lhs) - numpy.asarray(rhs)) / numpy.asarray(reference)

    def __init__(
        self,
        context: "System",
        equation: str,
        name: Optional[str] = None,
        reference: Union[Number, Collection, numpy.ndarray, str] = 1,
    ):
        """Initialization parameters:
        ----------
        - context : cosapp.systems.System
            CoSApp System to which this residue is linked
        - equation : str
            Equation of the kind 'hls == 'rhs', defining residue lhs - rhs.
        - name : str or None, optional
            Residue name; default None (built from equation)
        - reference : Number, numpy.ndarray or "norm", optional
            Reference value(s) used to normalize the equation; default is 1.
            If value is "norm", actual reference value is estimated from order of magnitude.
        """
        check_arg(equation, 'equation', str)
        check_arg(reference, 'reference', (Number, Collection, numpy.ndarray, str),
            lambda r: r == "norm" if isinstance(r, str) else numpy.all(numpy.asarray(r) > 0)
        )
        super().__init__(context, "temp")
        self.__sides = None   # type: EvalString
        self.__equation = ""  # type: str
        self.__set_equation(equation)
        self._name = name or self.equation

        if isinstance(reference, str) and reference == "norm":  # type check avoids a warning for numpy arrays
            reference = Residue.residue_norm(*self.eval_sides())
        self.reference = reference
        self.update()

    @property
    def equation(self) -> str:
        """str: Equation defining the residue"""
        return self.__equation

    @staticmethod
    def split_equation(equation: str) -> Tuple[str, str]:
        """Parses and splits an equation of the kind 'lhs == rhs'.
        Returns string tuple (lhs, rhs).
        """
        eqsign = "=="
        check_arg(equation, 'equation', str,
            lambda s: s.count(eqsign) == 1 and s.replace(eqsign, '').count('=') == 0)
        lhs, rhs = equation.split(eqsign)
        lhs, rhs = lhs.strip(), rhs.strip()
        if len(lhs) == 0 or len(rhs) == 0:
            raise SyntaxError(f"Equation should be of the kind 'lhs == rhs'; got {equation!r}")
        return lhs, rhs

    def __set_equation(self, equation: str) -> None:
        """Checks that the two sides of an equation of the kind 'lhs == rhs' are compatible,
        and that residue (lhs - rhs) is not trivially constant.
        Left- and right-hand sides are finally stored as two evaluable expressions (EvalString).
        """
        context = self.context
        lhs, rhs = Residue.split_equation(natural_varname(equation))
        sides = EvalString(f"({lhs}, {rhs})", context)
        if sides.constant:
            raise RuntimeWarning(f"Equation {lhs} == {rhs} is trivially constant")
        lval, rval = sides.eval()
        try:
            lval == rval
        except:
            raise TypeError(f"Expressions {lhs!r} and {rhs!r} are not comparable")
        else:
            self.__sides = sides
            self.__equation = f"{lhs} == {rhs}"

    def __str__(self) -> str:
        name = self.__equation or self._name
        return f"{name} := {self._value}"

    def eval_sides(self) -> Tuple[Any, Any]:
        """Evaluate and return left- and right-hand sides as a tuple"""
        return self.__sides.eval()

    def update(self) -> Union[Number, numpy.ndarray]:
        """Update the residue value

        Returns
        -------
        Number or numpy.ndarray
            The updated residues
        """
        lval, rval = self.__sides.eval()
        self._value = Residue.evaluate_residue(
            lval, rval, self.reference
        )
        return self._value

    def copy(self) -> Residue:
        """Copy the residue object.

        Returns
        -------
        Residue
            The duplicated residue
        """
        return Residue(
            self.context, self.__equation, self.name, reference=self.reference
        )

    def to_dict(self) -> Dict[str, Any]:
        """Returns a JSONable representation of the equation.
        
        Returns
        -------
        Dict[str, Any]
            JSONable representation
        """
        ref = self.reference
        if isinstance(self.reference, numpy.ndarray):
            ref = self.reference.tolist()
        return {
            "context": self.context.contextual_name,
            "equation": self.__equation,
            "name": self.name,
            "reference": str(ref)
        }


class DeferredResidue:
    """Class representing a residue whose left-hand side evaluation is deferred.

    The right-hand side of the residue is a targetted quantity.
    Upon request, a Residue object is generated, with lhs = value(target).

    Parameters
    ----------
    - context : cosapp.systems.System
        CoSApp System to which residue is linked.
    - target : str
        Targetted quantity, and left-hand side of the equation.
    - reference: Number or numpy.ndarray or None
        Reference value to normalize the residue; default is unity.
    - variables: Set[str]
        Names of variables involved in the residue
    """
    def __init__(self, context: "System", target: str, reference=1.0):
        from cosapp.systems import System
        check_arg(context, "context", System)
        check_arg(target, "target", str)
        self.__context: System = context
        self.reference = reference
        self.target = target

    @property
    def context(self):
        """System: evaluation context of residue"""
        return self.__context

    @property
    def target(self) -> str:
        """str: targetted quantity"""
        return str(self.__lhs)

    @target.setter
    def target(self, target: str) -> None:
        """Set targetted expression"""
        lhs = EvalString(target, self.context)
        if lhs.constant:
            raise ValueError(f"Targetted expression {lhs!r} appears to be constant")
        self.__lhs = lhs
        self.__vars = lhs.variables()

    @property
    def variables(self) -> FrozenSet[str]:
        """FrozenSet[str]: names of variables involved in residue"""
        return self.__vars

    def target_value(self) -> Any:
        """Evaluates and returns current value of target"""
        return self.__lhs.eval()

    def equation(self) -> str:
        """Returns target equation with updated lhs value"""
        return f"{self.target} == {self.target_value()!r}"

    def make_residue(self, reference=None) -> Residue:
        """Generates the residue corresponding to equation 'target == value(target)'"""
        if reference is None:
            reference = self.reference
        return Residue(self.context, self.equation(), reference=reference)

    def __repr__(self) -> str:
        clsname = self.__class__.__name__
        return f"{clsname}({self.context.name}, {self.target}, reference={self.reference})"



"""Class definition for SurrogateModel, the base class for all surrogate models.
"""
import abc


class SurrogateModel(abc.ABC):
    """Abstract interface for surrogate models.
    """

    @abc.abstractmethod
    def train(self, x, y):
        """
        Train the surrogate model with the given set of inputs and outputs.

        Parameters
        ----------
        x : array-like
            Training input locations
        y : array-like
            Model responses at given inputs.
        """
        pass

    @abc.abstractmethod
    def predict(self, x):
        """
        Calculate a predicted value of the response based on the current trained model.

        Parameters
        ----------
        x : array-like
            Point(s) at which the surrogate is evaluated.
        """
        pass


class MultiFiSurrogateModel(SurrogateModel):
    """
    Base class for surrogate models using multi-fidelity training data.
    """

    def train(self, x, y):
        """
        Calculate a predicted value of the response based on the current trained model.

        Parameters
        ----------
        x : array-like
            Point(s) at which the surrogate is evaluated.
        y : array-like
            Model responses at given inputs.
        """
        self.train_multifi([x], [y])

    @abc.abstractmethod
    def train_multifi(self, x, y):
        """
        Train the surrogate model, based on the given multi-fidelity training data.

        Parameters
        ----------
        x : list of (m samples, n inputs) ndarrays
            Values representing the multi-fidelity training case inputs.
        y : list of ndarray
            output training values which corresponds to the multi-fidelity
            training case input given by x.
        """
        pass
